#include <math.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>

#include "../src/grind.h"

/*  externals */
extern int pargc;
extern char **pargv;
extern int mode_opt;
extern pthread_t gl_tid;
extern char *modelname;

void *xopen(void*);

/*  local pointers to graphics stack */
_grcall *first_of_stack, *last_of_stack;

/*  id's for (possibly) multiple windows */
int winh = 500, winw = 500;


void startgraphics(_grcall *f) {
  static int rc = -1;

  /*  set first element of stack */
  first_of_stack = last_of_stack = f;

  /*  if called first time start the openGL thread */
  if (!gl_tid) {
    if ( (rc = pthread_create(&gl_tid, NULL, xopen, NULL)) ) {
      printf("ERROR; return code from pthread_create() is %d\n", rc);
      exit(-1);
    }
  }
  else 
    return;
}

int checkgraphics() {
  return (int)gl_tid;
}

int init() {
  /* set up projection for 3D */
  setProjection();
  return 1;
}

const struct timespec delay = {0, 10000};

void idle() {
  boolean redraw = false;
  _grcall *tmp = NULL;
  /*  first draw new things */
  redraw = check4append(&tmp, &last_of_stack);
  if (redraw) {
    /*  we only need to draw the new graphics */
    drawstack(tmp, last_of_stack);
  }

  /*  check whether the graphics stack will be erased */
  if (checkErase(first_of_stack, &last_of_stack)) {
    /*  we need to redraw the entire stack */
    drawstack(first_of_stack, last_of_stack);
    return;
  }

  if (!redraw)
    /*  if nothing happened sleep a little bit */
    nanosleep(&delay, NULL);
}

int ExponentText(char *base, char *pow, double x, double y, double z) {
  int len = strlen(base);
  char *all = ccalloc(strlen(base)+strlen(pow)+1, "'all' in ExponentText");
  
  addGrcall(MOVE, NULL, x, y, z);
  addGrcall(TEXT, base, 0., 0., 0.);
  strncpy(all, "          ", len);
  strcpy(&(all[len]), pow);
  addGrcall(MOVE, NULL, x, y + .015, z);
  addGrcall(TEXT, all, 0., 0., 0.);
  return 1;
}
