#include <math.h>
#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>

#include "../graphics/graphics.h"

int xclose();

extern _grcall *first_of_stack, *last_of_stack;

static int scale = 500;
static int width, height, lastx, lasty, newx, newy;
static double margin = 0.3;
static int twod = 1;

static Display *xdisplay;
char *server = NULL;
static Cursor cursor;
static Window window;
GC gcline;
XEvent event;
KeySym key;
XSizeHints hint;
XWindowAttributes attributes;
XSetWindowAttributes setattributes;
int screen,depth,class,colorscreen;
Visual *visual;
static unsigned long white, black, foreground, background;
XPoint *points;
Colormap new_colormap;
static XColor *colors;

int triplt(double,double,double,double*, double*);
int setProjection(); 
int init();                             /* ./graphics.c */
void idle();                            /* ./graphics.c */
int writemousecoord(int, int, int, int);/* ../src/mutex.c */
int xmainloop();

int xpos(double x)
{
  return (int)((margin+x)/(2.*margin + 1.)*scale + 0.5);
}

int ypos(double x)
{
  return height-(int)((margin+x)/(2.*margin + 1.)*scale + 0.5);
}

int TriPlt(double *coord) {
  double x, y;
  if (!twod) 
      triplt(coord[0],coord[1],coord[2],&x, &y);
  else {
    x = coord[0];
    y = coord[1];
  }
  newx = xpos(x);
  newy = ypos(y);
  return 1;
}

int xCall(const _grcall *tmp_grcall) {
  int i=0, type;
  char *str;
  double *coord;

  switch (type = tmp_grcall->type) {
  case CLEAR:  
    XClearWindow(xdisplay,window);
    break;
  case EXIT:
    xclose();
    pthread_exit(NULL);
    break;
  case TWOD:
    twod = 1;
    break;
  case THREED:
    twod = 0;
    break;
  case GR_POINT:
  case MOVE:
  case DRAW:
    coord = (double*)(tmp_grcall->val);
    TriPlt(coord);
    if (type == DRAW) {
      XDrawLine(xdisplay,window,gcline,lastx,lasty,newx,newy);
      lastx = newx;
      lasty = newy;
    }
    else {
      lastx = newx;
      lasty = newy;
      if (GR_POINT == type) XDrawPoint(xdisplay,window,gcline,lastx,lasty);
    }
    break;
  case ORECT:  
  case FRECT:
    coord = (double*)(tmp_grcall->val);
    TriPlt(coord);
    lastx = newx;
    lasty = newy;
    if (ORECT == type)
      XDrawRectangle(xdisplay,window,gcline,lastx-scale/100,lasty-scale/100,scale/50,scale/50);
    else
      XFillRectangle(xdisplay,window,gcline,lastx-scale/100,lasty-scale/100,scale/50,scale/50);
    break;
  case TEXT:
    str = (char*)(tmp_grcall->val);
    XDrawString(xdisplay, window, gcline, lastx, lasty,str,strlen(str));
    break;  
  case COLOR:
    i = *(int*)(tmp_grcall->val);
    if (colorscreen) {
      i = colorWrap(i);
      XSetForeground(xdisplay, gcline, colors[i].pixel);
    }
    break;
  default:
    fprintf(stderr,"error: trying to draw unknown primitive: %d\n", 
	    tmp_grcall->type);
    break;
  }
  return 1;
}

void MakeColorMap()
{
  int i,colormap_size;

  colormap_size = DisplayCells(xdisplay, screen);
  if ((colors = (XColor *)calloc(colormap_size,sizeof(XColor))) == NULL) {
    fprintf(stderr, "No memory for setting up colormap\n");
    exit(1);
  }
  for (i=0; i < colormap_size; i++) {
    colors[i].pixel = i;
    colors[i].flags = DoRed | DoGreen | DoBlue;
  }
  XQueryColors(xdisplay,DefaultColormap(xdisplay,screen),
	       colors,colormap_size); 
  for (i=0; i < NCOLORS; i++)
    if (i < colormap_size) {
      colors[i].red = 255*257*colormap[i][0];
      colors[i].green = 255*257*colormap[i][1];
      colors[i].blue = 255*257*colormap[i][2];
    } 
    else fprintf(stderr,"Cannot set color %d, your screen has colormap size %d\n",i,colormap_size);

  new_colormap = XDefaultColormap(xdisplay,screen);
  for (i=0; i < NCOLORS && i < colormap_size; i++)
    XAllocColor(xdisplay,new_colormap,&colors[i]);
}

void *xopen(void *np) {
  extern char **pargv;
  extern char *modelname;

  xdisplay = XOpenDisplay(server);
  if (!xdisplay) {
    printf("Failed to open display connection to %s\n",
      XDisplayName(server));
    return(0);
  }
  screen = DefaultScreen(xdisplay);
  depth = XDefaultDepth(xdisplay, screen);
  visual = XDefaultVisual(xdisplay, screen);
  class = visual->class;
  if (depth == 1) colorscreen = 0;
  else if (class == GrayScale) colorscreen = 0;
  else if (class == PseudoColor) colorscreen = 1;
  else if (class == DirectColor) colorscreen = 1;
  else if (class == TrueColor) colorscreen = 1;
  else if (class == StaticColor) colorscreen = 1;
  else if (class == StaticGray) colorscreen = 1;
  black = BlackPixel(xdisplay, screen);
  white = WhitePixel(xdisplay, screen);
  if (colorscreen) {
    background = black;
    foreground = white;
  } else {
    background = white;
    foreground = black;
  }
  hint.x = 300; hint.y = 300;
  hint.width = scale; hint.height = scale;
  hint.flags = PPosition | PSize;

  MakeColorMap(); 
  setattributes.colormap = new_colormap; 
  setattributes.background_pixel = BlackPixel(xdisplay,screen);
  setattributes.border_pixel = WhitePixel(xdisplay,screen);
  window = XCreateSimpleWindow(xdisplay,
    DefaultRootWindow(xdisplay),
    hint.x, hint.y, hint.width, hint.height, 5,
    foreground, background);
  XSetStandardProperties(xdisplay, window, modelname, modelname,
    None, pargv, 0, &hint);
  XGetWindowAttributes(xdisplay,window,&attributes);
  width = attributes.width;
  height = attributes.height;
  gcline = XCreateGC(xdisplay, window, 0, 0);
  XSetBackground(xdisplay, gcline, background);
  XSetForeground(xdisplay, gcline, foreground);
  cursor = XCreateFontCursor(xdisplay, XC_arrow);
  XDefineCursor(xdisplay,window,cursor);
  XSelectInput(xdisplay, window, ButtonPressMask | ExposureMask | KeyPressMask);
  XMapRaised(xdisplay, window);
  XNextEvent(xdisplay, &event); 
  init();
  while(1)
    xmainloop();
  pthread_exit(NULL);
}

int drawstack(const _grcall *first, const _grcall *last) {
  const _grcall *tmp = first;

  xCall(tmp);
  while(tmp != last) {
    tmp = tmp->next;
    xCall(tmp);
  } 
  return 1;
}

int xmainloop() {
  int ix, iy;
  char string[20];
  int count, nchar = 20;
  KeySym key;
  XComposeStatus cs;
  while (XEventsQueued(xdisplay,QueuedAfterFlush) > 0) {
    XNextEvent(xdisplay, &event);
    switch (event.type) {
      case Expose:
        XGetWindowAttributes(xdisplay,window,&attributes);
        width = attributes.width;
        height = attributes.height;
	scale = (width > height) ? height : width;
	drawstack(first_of_stack, last_of_stack);
        break;
      case ButtonPress:
	ix = event.xbutton.x;
	iy = height - event.xbutton.y;
	writemousecoord(0, 0, ix, iy);
        break;
      case KeyPress:
        count = XLookupString(&event.xkey, string, nchar, &key, &cs);
        break;
    }
  }
  idle();
  return 1;
}

int getrealcoord(int i, int j, double *x, double *y, double *z) { 
  *x = 1.6*i/scale - 0.3;
  *y = 1.6*j/scale - 0.3;
  return 1;
}

int xclose() {
  XFreeGC(xdisplay, gcline);
  XDestroyWindow(xdisplay, window);
  XCloseDisplay(xdisplay);
  return 1;
}

