#include <math.h>
#include <string.h>
#include <pthread.h>

#include "../src/grind.h"

#define WRDLN 256

static const double ticklen = 0.02;
static const double xlabeldist = 0.2;
static const double ylabeldist = 0.2;
static const double zlabeldist = 0.3;
static const double xticklabeldist = 0.1;
static const double yticklabeldist = 0.1;
static const double zticklabeldist = 0.2;

extern int mode_opt;
extern double finish;
extern int nintervals;

extern pthread_t gl_tid;


int num2text(double, double, double, double);
double max(double,double);

double val2ax(double val, int ax) {
  /* return value in ax-coordinates [0..1] given axis and value of the 
     var/par/expr on that axis 
  */
  return ( ( axis[ax].log ? ( (val > TINY) ? log10(val) : -9.) : val )
	   - axis[ax].min)
    / (axis[ax].max-axis[ax].min);
}

double ax2val(double val, int ax) {
  /* return value of var/par/expr on axis 'ax' given the location 
     in the ax-coordinates [0..1]
   */
  if (axis[ax].log && axis[ax].min > axis[ax].max)
    printf("old fortran-GRIND does something strange in this case\n");

  if (axis[ax].log)
    return pow(10., (val * (axis[ax].max-axis[ax].min) + axis[ax].min));
  else
    return val * (axis[ax].max-axis[ax].min) + axis[ax].min;
}

boolean storeaxis(boolean store) {
  /* store values of var/par/expr's associated with each of the axis in 
     local, static variables ('store' is TRUE)
     reset values of var/par/expr's associated with each of the axis to
     values stored previously in local variables
     note that for variables the 'istate' is stored and restored
   */
  static double oldvals[3];

  if (store) {
    if (axis[0].var<nvar) oldvals[0] = istate[axis[0].var];
    else oldvals[0] = all_values[axis[0].var];
    if (axis[1].var<nvar) oldvals[1] = istate[axis[1].var];
    else oldvals[1] = all_values[axis[1].var];
    if (axis[2].var<nvar) oldvals[2] = istate[axis[2].var];
    else oldvals[2] = all_values[axis[2].var];
  }
  else {
    if (axis[0].var<nvar) istate[axis[0].var] = oldvals[0];
    else all_values[axis[0].var] = oldvals[0];
    if (axis[1].var<nvar) istate[axis[1].var] = oldvals[1];
    else all_values[axis[1].var] = oldvals[1];
    if (axis[2].var<nvar) istate[axis[2].var] = oldvals[2];
    else all_values[axis[2].var] = oldvals[2];
  }
  return store;
}

int crd(double *xyz, const double *yt) {
  /* for each axis store in xyz[] the corresponding value in yt[] if  */
  /* a variable is associated with that axis, otherwise store the value */
  /* of the parameter / expression associated with that axis */
  /* obvioulsy, yt[] is a state vector */
  xyz[0] = val2ax((axis[0].var < nvar ?  /* is a var or a par/expr displayed */
		   yt[axis[0].var] :     /* return corresponding yt[] value */
		   all_values[axis[0].var]),/*otherwise return par/expr value*/
		  0);                       /* of 0-th axis */
  xyz[1] = val2ax((axis[1].var < nvar ?     /* idem */
		   yt[axis[1].var] : 
		   all_values[axis[1].var]), 
		  1);
  if (DIM3 != mode_opt)                     /*  check 2D vs. 3D */
    xyz[2] = 0.;
  else
    xyz[2] = val2ax((axis[2].var < nvar ? 
		     yt[axis[2].var] : 
		     all_values[axis[2].var]), 
		    2);
  return 1;
}

int setax(double *yt, double val, int iax) {
  /* yt[] is a state vector. Set it to ax2val('val') if 'iax' specifies an 
     axis with a variable associated with it. Otherwise set the par/expr
     associated with that axis
   */
  if (axis[iax].var < nvar)
    yt[axis[iax].var] = ax2val(val, iax);
  else
    all_values[axis[iax].var] = ax2val(val, iax);
  return 1;
}

static double cosa = 0., sina = 0.;
#define rx(a, b) (a*cosa - b*sina)
#define ry(a, b) (a*sina + b*cosa)
int tr_size = 1;
int triangle(double *pop, double *dpop) {
  /* draw a triangle; code directly based on fortran-grind
   */
  static double *yt = NULL;
  double xyz[3], nxyz[3];
  double dx, dy, h;
  int i;
  double size = 0.01 *(double)tr_size;
  int col;

  if (!yt) yt = dcalloc(nvar, "'yt' in triangle()");

  crd(xyz, pop); /* store 'pop' in 'yt[]' if there're var's on the axis */
  for (i=0; i<nvar; i++)
    yt[i] = pop[i] + dpop[i];
  crd(nxyz, yt); 

  dx = nxyz[0] - xyz[0];
  dy = nxyz[1] - xyz[1];
  h = max(sqrt(dx*dx+dy*dy), 1e-9);
  cosa = dx/h;
  sina = dy/h;
  col = WHITE;
  addGrcall(COLOR, (void*)&col, 0., 0., 0.);
  addGrcall(MOVE, NULL, xyz[0], xyz[1], xyz[2]);
  addGrcall(DRAW, NULL, 
	    xyz[0]+rx(0.0, -size), 
	    xyz[1]+ry(0.0, -size),
	    xyz[2]);
  addGrcall(DRAW, NULL, 
	    xyz[0]+rx(3.0*size, 0.0), 
	    xyz[1]+ry(3.0*size, 0.0), 
	    xyz[2]);
  addGrcall(DRAW, NULL, 
	    xyz[0]+rx(0.0, size), 
	    xyz[1]+ry(0.0, size), 
	    xyz[2]);
  addGrcall(DRAW, NULL, xyz[0], xyz[1], xyz[2]);
    
  return 1;
}
/* clean up 'rx' and 'ry' */
#undef rx
#undef ry

int displayPosition(double *yt) {
  /* draw lines perpendicular on axis, indicating the state vector 'yt[]'
   */
  double x, y, z;
  int col;

  x = val2ax((axis[0].var < nvar ? 
	      yt[axis[0].var] : 
	      all_values[axis[0].var]), 0);
  y = val2ax((axis[1].var < nvar ? 
	      yt[axis[1].var] : 
	      all_values[axis[1].var]), 1);
  if (DIM3 == mode_opt)
    z = val2ax((axis[2].var < nvar ? 
		yt[axis[2].var] : 
		all_values[axis[2].var]), 2);
  else
    z = 0.;

  col = WHITE;
  addGrcall(COLOR, (void*)&col, 0., 0., 0.);
  addGrcall(MOVE, NULL, x, 0, z);
  addGrcall(DRAW, NULL, x, y, z);
  addGrcall(MOVE, NULL, 0, y, z);
  addGrcall(DRAW, NULL, x, y, z);
  return 1;
}

int setMark(const double *yt, boolean stable) {
  /* draw a small rectangle in state-space at position 'xyz'
     rect is filled/open depending on 'stable'
   */
  double xyz[3];
  int col;

  crd(xyz, yt); /* store 'yt[]' in xyz if there's var's on the axes */

  if (stable) {
    col = RED;
    addGrcall(COLOR, (void*)&col, 0., 0., 0.);
    addGrcall(FRECT, NULL, xyz[0], xyz[1], xyz[2]);
  }
  else {
    col = WHITE;
    addGrcall(COLOR, (void*)&col, 0., 0., 0.);
    addGrcall(ORECT, NULL, xyz[0], xyz[1], xyz[2]);
  }
  return 1;
}

int start_continuation(boolean stable) {
  /* initialization for drawing a continuation-line; set color depending 
     on 'stable', move to the first position
   */
  double x, y, z;
  int col;

  if (stable) {
    col = BLUEISH;
    addGrcall(COLOR, (void*)&col, 0., 0., 0.);
  }
  else {
    col = WHITE;
    addGrcall(COLOR, (void*)&col, 0., 0., 0.);
  }

  /*  for all three axes take the value of the associated parameter  */
  /*  or expression or parameter */
  x = val2ax(all_values[axis[0].var], 0);
  y = val2ax(all_values[axis[1].var], 1);
  if (mode_opt == DIM3)
    z = val2ax(all_values[axis[2].var], 2);
  else
    z = 0.;
  /*  move to the first position of the trajectory */
  return addGrcall(MOVE, NULL, x, y, z);
}

int start_trajectory(int c) {
  /* initialization for drawing a trajectory; set color to 'c', depending on 
     'points_opt' either draw a POINT or move to first position
   */
  double x, y, z;
  
  addGrcall(COLOR, (void*)&c, 0., 0., 0.);

  /*  for all three axes take the value of the associated parameter  */
  /*  or expression or parameter */
  x = val2ax(all_values[axis[0].var], 0);
  y = val2ax(all_values[axis[1].var], 1);
  if (mode_opt == DIM3)
    z = val2ax(all_values[axis[2].var], 2);
  else
    z = 0.;
  if (points_opt)
    /*  set point at first position of the trajectory */
    return addGrcall(GR_POINT, NULL, x, y, z);
  else
    /*  move to the first position of the trajectory */
    return addGrcall(MOVE, NULL, x, y, z);
}

int add2trajectory() {
  /* add to an existing trajectory or continuation line
   */
  double x, y, z;

  /*  for all three axes take the value of the associated parameter  */
  /*  or expression or parameter */
  x = val2ax(all_values[axis[0].var], 0);
  y = val2ax(all_values[axis[1].var], 1);
  if (mode_opt == DIM3)
    z = val2ax(all_values[axis[2].var], 2);
  else
    z = 0.;
  if (points_opt)
    /*  set point at next position of the trajectory */
    return addGrcall(GR_POINT, NULL, x, y, z);
  else
    /*  draw line piece to this next position */
    return addGrcall(DRAW, NULL, x, y, z);
}

int drawTimeSeries(boolean *toreport) {
  /* a run is drawn as timeplot. 
     var/par/expr's in 'toreport' are drawn.
     the run is drawn from ax-min to ax-max; this is possibly different 
     from t=0 and t=t_max (this complicates code a bit).
     therefor we calculate begin-time = tb, end-time = te
   */
  const int v_ax = 4;
  const int h_ax = 5;
  int i, t, tb, te, c;
  double finint;
  finint = finish/nintervals;
  if (ax2val(0., h_ax) >= 0.) /*  the time axis starts at some value > 0. */
    /*  start reading the data buffer at that time, adjusted by finint */
    tb = (int)(ax2val(0., h_ax) / finint);
  else
    /*  start reading the buffer at 0 */
    tb = 0;
  /*  read until the end, adjusted by finint */
  te = (int)(0.5 + ax2val(1., h_ax)/finint);

  if(rundatalength < tb)
    /*  the start of the axis is beyond the scope of the buffer */
    return 0;
  if(rundatalength < te)
    /*  the axis continues beyond the buffer; adjust te accordingly */
    te = rundatalength - 1;

  for (i=0; i<nnames; i++) {
    /*  color is RED or higher */
    c = i+1;
    if (outputbuf[i] && toreport[i]) {
      /* only if this var/par/expr is set in both buffers it is plotted */
      addGrcall(COLOR, (void*)&c, 0., 0., 0.);
      /*  move to first point */
      addGrcall(MOVE, NULL, 
		val2ax(rundata[0][tb],h_ax), 
		val2ax(rundata[i+1][tb], v_ax), 0.);
      /*  draw the rest of the buffer */
      for (t=tb+1; t<=te; t++)
	addGrcall(DRAW, NULL, 
		  val2ax(rundata[0][t],h_ax), 
		  val2ax(rundata[i+1][t], v_ax), 0.);
      if (label_opt) {
	/*  draw the names of the variables in color and at proper locations */
	addGrcall(MOVE, NULL, 
		  1.+yticklabeldist, 
		  val2ax(rundata[i+1][te], v_ax), 0.);
	addGrcall(TEXT, (char*)(all_names[i]), 0., 0., 0.);
      }
    }
  }
  return 1;
}

int Draw2Daxis() {
  const int h_ax = 0;
  const int v_ax = 1;
  int i, t, col;
  double x, dx;

  /*  setup drawing in 2d space */
  addGrcall(TWOD, NULL, 0., 0., 0.);
  /*  set color */
  col = WHITE;
  addGrcall(COLOR, (void*)&col, 0., 0., 0.);
  /*  4 lines spanning state space */
  addGrcall(MOVE, NULL, 0., 0., 0.);
  addGrcall(DRAW, NULL, 1., 0., 0.);
  addGrcall(DRAW, NULL, 1., 1., 0.);
  addGrcall(DRAW, NULL, 0., 1., 0.);
  addGrcall(DRAW, NULL, 0., 0., 0.);

  if (label_opt) {
  /*  x-axis label */
  addGrcall(MOVE, NULL, .8, -xlabeldist, 0.);
  addGrcall(TEXT, axis[h_ax].label, 0., 0., 0.);

  /*  distance between x-axis tick marks and labels */
  t = axis[h_ax].ticks;
  /*  place the ticks and corresponding labels */
  for (i=0, x=0., dx=1./(t-1.); i<t; i++, x += dx) {
    /*  the tick */
    addGrcall(MOVE, NULL, x, 0., 0.);
    addGrcall(DRAW, NULL, x, -ticklen, 0.);
    /*  the label */
    num2text(ax2val(x, h_ax), x, -xticklabeldist, 0.);
  }

  /*  y-axis label */
  addGrcall(MOVE, NULL, -ylabeldist, 0.8, 0.);
  addGrcall(TEXT, axis[v_ax].label, 0., 0., 0.);

  /*  distance between y-axis tick marks and labels */
  t = axis[v_ax].ticks;
  /*  place the ticks and corresponding labels */
  for (i=0, x=0., dx=1./(t-1.); i<t; i++, x += dx) {
    /*  the tick */
    addGrcall(MOVE, NULL, 0., x, 0.);
    addGrcall(DRAW, NULL, -ticklen, x, 0.);
    /*  the label */
    num2text(ax2val(x, v_ax), -yticklabeldist, x, 0.);
  }
  }
  return 1;
}

int Draw3Daxis() {
  const int h_ax = 0;
  const int v_ax = 1;
  const int d_ax = 2;
  int i, t, col;
  double x, dx;

  /*  setup drawing in 3d space */
  addGrcall(THREED, NULL, 0., 0., 0.);
  /*  set color */
  col = WHITE;
  addGrcall(COLOR, (void*)&col, 0., 0., 0.);
  /*  9 lines spanning x-y-z state space */
  addGrcall(MOVE, NULL, 0., 0., 0.);
  addGrcall(DRAW, NULL, 1., 0., 0.);
  addGrcall(DRAW, NULL, 1., 1., 0.);
  addGrcall(DRAW, NULL, 0., 1., 0.);
  addGrcall(DRAW, NULL, 0., 0., 0.);
  addGrcall(DRAW, NULL, 0., 0., 1.);
  addGrcall(DRAW, NULL, 1., 0., 1.);
  addGrcall(DRAW, NULL, 1., 0., 0.);
  addGrcall(MOVE, NULL, 0., 0., 1.);
  addGrcall(DRAW, NULL, 0., 1., 1.);
  addGrcall(DRAW, NULL, 0., 1., 0.);
  if (closed_opt) {
    addGrcall(MOVE, NULL, 0., 1., 1.);
    addGrcall(DRAW, NULL, 1., 1., 1.);
    addGrcall(DRAW, NULL, 1., 1., 0.);
    addGrcall(MOVE, NULL, 1., 1., 1.);
    addGrcall(DRAW, NULL, 1., 0., 1.);
  }
  if (label_opt) {
  /*  x-axis label */
  addGrcall(MOVE, NULL, .8, -xlabeldist, 1.);
  addGrcall(TEXT, axis[0].label, 0., 0., 0.);
  /*  distance between x-axis tick marks and labels */
  t = axis[h_ax].ticks;
  /*  place the ticks and corresponding labels */
  for (i=0, x=0., dx=1./(t-1.); i<t; i++, x += dx) {
    /*  the tick */
    addGrcall(MOVE, NULL, x, 0., 1.);
    addGrcall(DRAW, NULL, x, -ticklen, 1.);
    /*  the label */
    num2text(ax2val(x, h_ax), x, -xticklabeldist, 1.);
  }

  /*  y-axis label */
  addGrcall(MOVE, NULL, 1.+ylabeldist, .8, 0.);
  addGrcall(TEXT, axis[1].label, 0., 0., 0.);
  /*  distance between y-axis tick marks and labels */
  t = axis[v_ax].ticks;
  /*  place the ticks and corresponding labels */
  for (i=0, x=0., dx=1./(t-1.); i<t; i++, x += dx) {
    /*  the tick */
    addGrcall(MOVE, NULL, 1., x, 0.);
    addGrcall(DRAW, NULL, 1. + ticklen, x, 0.);
    /*  the label */
    num2text(ax2val(x, v_ax), 1. + yticklabeldist, x, 0.);
  }

  /*  z-axis label */
  addGrcall(MOVE, NULL, -zlabeldist, 1., .7);
  addGrcall(TEXT, axis[d_ax].label, 0., 0., 0.);
  /*  distance between z-axis tick marks and labels */
  t = axis[d_ax].ticks;
  /*  place the ticks and corresponding labels */
  for (i=0, x=0., dx=1./(t-1.); i<t; i++, x += dx) {
    /*  the tick */
    addGrcall(MOVE, NULL, 0., 1., x);
    addGrcall(DRAW, NULL, -ticklen, 1., x);
    /*  the label */
    num2text(ax2val(x, d_ax), -zticklabeldist, 1., x);
  }
  }
  return 1;
}

int drawTimeAxis() {
  const int h_ax = 5;
  const int v_ax = 4;
  int i, t, col;
  double x, dx;

  /*  setup drawing in 2d space */
  addGrcall(TWOD, NULL, 0., 0., 0.);
  /*  set color */
  col = WHITE;
  addGrcall(COLOR, (void*)&col, 0., 0., 0.);
  /*  4 lines spanning time plot */
  addGrcall(MOVE, NULL, 0., 0., 0.);
  addGrcall(DRAW, NULL, 1., 0., 0.);
  addGrcall(DRAW, NULL, 1., 1., 0.);
  addGrcall(DRAW, NULL, 0., 1., 0.);
  addGrcall(DRAW, NULL, 0., 0., 0.);
  if (label_opt) {
  /*  horizontal-axis label */
  addGrcall(MOVE, NULL, .8, -xlabeldist, 0.);
  addGrcall(TEXT, axis[h_ax].label, 0., 0., 0.);

  /*  distance between hortizontal-axis tick marks and labels */
  t = axis[h_ax].ticks;
  /*  place the ticks and corresponding labels */
  for (i=0, x=0., dx=1./(t-1.); i<t; i++, x += dx) {
    /*  the tick */
    addGrcall(MOVE, NULL, x, 0., 0.);
    addGrcall(DRAW, NULL, x, -ticklen, 0.);
    /*  the label */
    num2text(ax2val(x, h_ax), x, -xticklabeldist, 0.);
  }

  /*  distance between vertical-axis tick marks and labels */
  t = axis[v_ax].ticks;
  /*  place the ticks and corresponding labels */
  for (i=0, x=0., dx=1./(t-1.); i<t; i++, x += dx) {
    /*  the tick */
    addGrcall(MOVE, NULL, 0., x, 0.);
    addGrcall(DRAW, NULL, -ticklen, x, 0.);
    /*  the label */
    num2text(ax2val(x, v_ax), -yticklabeldist, x, 0.);
  }
  }
  return 1;
}

int DrawTakens(int delta) {
  int i, v, col;
  char text[16];
  double x, dx;
  int t, c;
  extern int nintervals;

  if (!outputbuf[axis[T_AXIS].var])
    return gr_warning("No corresponding data in output buffer\n");
  if ((int)(finish + .5) != nintervals)
    return gr_warning("Takens requires every timestep; make finish equal to nintervals\n");


  /*  setup drawing in 2d space */
  addGrcall(TWOD, NULL, 0., 0., 0.);
  /*  set color */
  col = WHITE;
  addGrcall(COLOR, (void*)&col, 0., 0., 0.);
  /*  4 lines spanning state space */
  addGrcall(MOVE, NULL, 0., 0., 0.);
  addGrcall(DRAW, NULL, 1., 0., 0.);
  addGrcall(DRAW, NULL, 1., 1., 0.);
  addGrcall(DRAW, NULL, 0., 1., 0.);
  addGrcall(DRAW, NULL, 0., 0., 0.);
  if (label_opt) {
  /*  t-axis label */
  snprintf(text, 15, "%s-%d", axis[T_AXIS].label, delta);
  addGrcall(MOVE, NULL, .8, -xlabeldist, 0.);
  addGrcall(TEXT, text, 0., 0., 0.);

  /*  distance between x-axis tick marks and labels */
  t = axis[T_AXIS].ticks;
  /*  place the ticks and corresponding labels */
  for (i=0, x=0., dx=1./(t-1.); i<t; i++, x += dx) {
    /*  the tick */
    addGrcall(MOVE, NULL, x, 0., 0.);
    addGrcall(DRAW, NULL, x, -ticklen, 0.);
    /*  the label */
    num2text(ax2val(x, T_AXIS), x, -xticklabeldist, 0.);
  }

  /*  y-axis label */
  addGrcall(MOVE, NULL, -ylabeldist, 0.8, 0.);
  addGrcall(TEXT, axis[T_AXIS].label, 0., 0., 0.);

  /*  distance between y-axis tick marks and labels */
  t = axis[T_AXIS].ticks;
  /*  place the ticks and corresponding labels */
  for (i=0, x=0., dx=1./(t-1.); i<t; i++, x += dx) {
    /*  the tick */
    addGrcall(MOVE, NULL, 0., x, 0.);
    addGrcall(DRAW, NULL, -ticklen, x, 0.);
    /*  the label */
    num2text(ax2val(x, T_AXIS), -yticklabeldist, x, 0.);
  }
  }
  /* work: */
  c = RED;
  v = 1 + axis[T_AXIS].var;
  addGrcall(COLOR, (void*)&c, 0., 0., 0.);
  /*  draw the rest of the buffer */
  for (t=delta; t<=nintervals && !Interrupt(); t++)
    addGrcall(GR_POINT, NULL, 
		  val2ax(rundata[v][t-delta],T_AXIS), 
		  val2ax(rundata[v][t], T_AXIS), 0.);
  return !Interrupt();
}
  

boolean outside(double a,double b,double c) {
  /* check whether the arguments lie outside the state-space
   */

  if (a < ax2val(0.0, X_AXIS)-TINY || a > ax2val(1.0, X_AXIS)+TINY ||
      b < ax2val(0.0, Y_AXIS)-TINY || b > ax2val(1.0, Y_AXIS)+TINY ||
      ((DIM3 == mode_opt) && 
       (c < ax2val(0.0, Z_AXIS)-TINY || c > ax2val(1.0, Z_AXIS)+TINY)))
    return true;
  
  return false;
} 


int num2text(double val, double x, double y, double z) {
  /* format a text string for base and exponent of a value.
     plot it at location x,y,z.
   */
  char text[12], base[12], *pow;
  int exp;

  snprintf(text, 11, "%.3g", val);
  pow = strchr(text, 'e');
  if (pow) {
    /*  pointer arithmatic; 'exp' is index in text where exponent starts */
    exp = pow - text;
    if (1 == exp && '1' == text[0])
      /*  if base equals "1" only put "10" in base */
      strcpy(base, "10");
    else {
      /*  if base not equal to "1" concat base with "x10" */
      strncpy(base, text, exp);
      strcpy(base+exp, "x10");
    }
    pow++;
    if ('+' == pow[0])
      pow++;
    if ('0' == pow[0])
      pow++;
    ExponentText(base, pow, x, y, z);
  }
  else {
    strcpy(base, text);
    addGrcall(MOVE, NULL, x, y, z);
    addGrcall(TEXT, base, 0., 0., 0.);
  }


  return 1;
}

int closegraphics() {
  addGrcall(EXIT, NULL, 0., 0., 0.);
  if (gl_tid) {
    pthread_join(gl_tid, NULL);
  }
  gl_tid = 0;
  return 1;
}
