#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "pp.h"

EQN *var_stack = NULL;
EQN *exp_stack = NULL;
NODEP *dffeqns, *asignms;
NODEP cptree(),rmtree(),pasteim(),cutim(),deriv(),cltree();
NAMEP linlist = NULL;
char **varnms, **ivrnms;
char warnings[21][80];
int power = 0;
int nvar = 0, nimvar = 0, nnames = 0, jacob = 1, len_list = 0;
int nfatal =0, nwarning = 0, nlines = 1, fileOpen = 0;

FILE *src,*input;

int yyparse();
int wrhead();
int wrfcn();
int wrjac();
int wrmain();
int vdelete(char*);
int prtree(NODEP, FILE*);

int push(NODEP node, char *text, int var) 
{
  EQN *new;  
  new = (EQN *)malloc((unsigned)sizeof(EQN));
  if (new == NULL) {printf("grindpp: out of memory\n"); exit(0);}
  new->node = node;  
  if (strlen(text) > 10) {
    fprintf(stderr,"%s truncated to 10 characters\n",text);
    text[10] = '\0';
  }
  strncpy(new->text,text,11);
  if (var) {
    new->next = var_stack;
    var_stack = new;
  }
  else {
    new->next = exp_stack;
    exp_stack = new;
  }
  return 1;
}

int main(int argc, char *argv[])
{
  int i,j;
  char c, *filename = NULL;
  EQN *eqn;
 
  for (i=1; i < argc; i++) {
    if (argv[i][0] == '-') {
      for (j=0; (c=argv[1][j]) != '\0'; j++)
	if (c == 'j') jacob = 0;
    } else {
      if (!fileOpen) {
	filename = argv[i];
	fileOpen = 1;
      }
      else 
	printf("Too many filenames: %s ignored\n",argv[i]);
    }
  }
  if (fileOpen) 
    input = fopen(filename,"r");	  
  else {
    printf("Fatal error: supply a filename\n");
    exit(1);
  }
  printf("\n");
  printf("\t\t\t ********************************************* \n"); 
  printf("\t\t\t *  GReat INtegrator Differential equations  * \n"); 
  printf("\t\t\t *  Rob J de Boer & Ludo Pagie (c) 1983-2005 * \n"); 
  printf("\t\t\t *  Theoretical Biology, Utrecht University  * \n"); 
  printf("\t\t\t *             the Netherlands               * \n"); 
  printf("\t\t\t ********************************************* \n"); 
  printf("\n\nG R I N D (2.15) model listing:\n\n");

  yyparse();

  eqn = var_stack;
  dffeqns = (NODEP *)malloc(nvar*sizeof(NODE *));  
  asignms = (NODEP *)malloc(nimvar*sizeof(NODE *)); 
  varnms = (char **)malloc(nvar*sizeof(char *));
  ivrnms = (char **)malloc(nimvar*sizeof(char *)); 
  for (i = nvar-1; i >= 0; i--) {
    dffeqns[i] = eqn->node;
    varnms[i] = eqn->text;
    eqn = eqn->next;
  }  
  eqn = exp_stack;
  for (i = nimvar-1; i >= 0; i--) {
    asignms[i] = eqn->node;
    ivrnms[i] = eqn->text;
    eqn = eqn->next;
  }

  if (!nfatal) {    
    i = strlen(filename);
    if (0 == strcmp(&filename[i-4],".grd") ||
	0 == strcmp(&filename[i-4],".GRD") )
      filename[i-4] = '\0';
    src = fopen(strcat(filename,".c"),"w");
    wrhead(); 
    wrfcn(); 
    wrjac();
    wrmain();
  }
  printf("\n");  

  if (0 && !jacob) printf("Jacobian was suppressed\n");
  if (nfatal) printf("%d Fatal error(s)\n",nfatal);
  if (nwarning) 
     for (i=0; i <nwarning; i++)
       printf("%s\n",warnings[i]);
  if (!nfatal) printf("Preprocessing done!\n");
  return(nfatal);
}

int fp(char *line)
{
  return fprintf(src,"%s\n",line);
}

int wrhead()
{
  NAMEP name;
  int i,j;
  for (j=0; j<nvar; j++)
    vdelete(varnms[j]);
  for (j=0; j<nimvar; j++)
    vdelete(ivrnms[j]);
  for (i=0,name=linlist; name!=NULL; name=name->right,i++);
  nnames = nvar + nimvar + i;
  fp("#include <stdio.h>\n#include <math.h>\n");
  fp("double min(double, double), max(double, double), sign(double, double), heaviside(double);");
  fprintf(src,"const int nvar = %d;\n",nvar);
  fprintf(src,"const int nexpr = %d;\n",nimvar);
  fprintf(src,"const int npar = %d;\n",nnames-nvar-nimvar);  
  fprintf(src,"const int nnames = %d;\n",nnames); 
  fprintf(src,"double all_values[%d];\n",nnames);
  fprintf(src,"double * const par_values = all_values + %d;\n", nvar);
  fprintf(src,"const char *all_names[%d] = {\"%s\"",nnames,varnms[0]);
  for (i=1; i<nvar; i++)
    fprintf(src,",\"%s\"",varnms[i]);
  for (i=0; i<nimvar; i++)
    fprintf(src,",\"%s\"",ivrnms[i]);
  for (name=linlist; name!=NULL; name=name->right)
    fprintf(src,",\"%s\"",name->text);
  fp("};");
  fp("extern int positive_opt, forward_opt;");
  if (jacob) fp("int jacob_opt = 1;\n");
  else fp("int jacob_opt = 0;\n");
  fprintf(src,"int Initialize(int, char**);\n");  
  fprintf(src,"int mkVarPos(double[]);\n");  
  fprintf(src,"int GrindMainLoop(int);\n");  
  fprintf(src,"\n");  
  fprintf(src,"int mine(int argc, char *argv[])\n");  
  fprintf(src,"{\n");
  fprintf(src,"  int i;\n");
  fprintf(src,"  for (i=0; i < argc; i++)\n");
  fprintf(src,"    printf(\"%%s\" ,argv[i]);\n");
  fprintf(src,"  return argc;\n");
  fprintf(src,"}\n\n");
  return 1;
}

int wrVarPar()
{
  int i;
  NAMEP name;
  for (i=0; i<nvar; i++) 
    fprintf(src,"%s_ = _x_[%d];\n",varnms[i],i);
  for (i=nimvar,name=linlist; name!=NULL; name=name->right,i++)
    fprintf(src,"%s_ = par_values[%d];\n",name->text,i);
  for (i=0; i<nimvar; i++) {
    fprintf(src,"par_values[%d] = %s_ = ",i,ivrnms[i]);
    prtree(asignms[i],src);
    fprintf(src,";\n");
  }
  return 1;
}

NAMEP locate(text,found)
char text[]; int *found;
{
  int i = -1;
  NAMEP name, lname;
  for (lname=name=linlist; name!=NULL && i<0; name=name->right) {
    i = strcmp(name->text,text); 
    lname = name;
  }
  if (i>0) lname = lname->left;
  *found = i;
  return (lname);
}

int mkname(text) char text[];
{
  int found;
  NAMEP name,newn;
  if (strlen(text) > 10) {
    fprintf(stderr,"%s truncated to 10 characters\n",text);
    text[10] = '\0';
  }
  name = locate(text,&found);
  if (found == 0) { return(0); }
  else {
    newn = (NAMEP)malloc((unsigned)sizeof(NAME));
    strncpy(newn->text, text, 11);
    if (name == NULL) {
      newn->left = newn->right = NULL;
      if (linlist != NULL) {
        linlist->left = newn;
        newn->right = linlist;
      }
      linlist = newn;
    } else {
        newn->left = name;
        newn->right = name->right;
        if (name->right != NULL)
          name->right->left = newn;
        name->right = newn;
    }
    return (len_list++);
  }
  return 0;
}

int vdelete(text) char text[];
{
  int found;
  NAMEP name;
  name = locate(text,&found);
  if (found == 0) { 
    if (name->right != NULL) name->right->left = name->left;
    if (name->left != NULL) name->left->right = name->right;
    else linlist = name->right;
    free((char *)name);
  }else{
    sprintf(warnings[nwarning],"Warning in general: %s never used",text);
    if (nwarning < 20) nwarning++;
  }
  return 1;
}

int wrDouble()
{
  int i; 
  NAMEP name;
  fprintf(src,"double %s_",varnms[0]);
  for (i=1; i < nvar; i++)
    fprintf(src,",%s_",varnms[i]);
  for (i=0; i < nimvar; i++)
    fprintf(src,",%s_",ivrnms[i]);
  for (name=linlist; name!=NULL; name=name->right)
   fprintf(src,",%s_",name->text);
  fp(";");
  return 1;
}

int wrfcn()
{
  int i;
  fp("void fcn(double _x_[], double _dx_[])\n{"); 
  wrDouble();
  fp("int _i_;"); 
  fp("if (positive_opt) mkVarPos(_x_);"); 
  wrVarPar();
  for (i=0; i<nvar; i++) {
      fprintf(src,"_dx_[%d]=",i);
      prtree(dffeqns[i],src);
      fp(";");
  }
  fprintf(src,"if (!forward_opt)\n  for (_i_=0; _i_<%d; _i_++)\n",nvar); 
  fp("    _dx_[_i_] = - _dx_[_i_];"); 
  fp("}\n");
  return 1;
}
 
int wrjac()
{  
  NODEP node;
  int i,j,n;
  fp("void jac(double _x_[], double _jac_[])\n{"); 
  if (!jacob)
    fp("  printf(\"Jacobian was suppressed\\n\");"); 
  else {  
    wrDouble();
    fp("int _i_;");
    wrVarPar();
    for (n=0; n<nvar; n++)
    for (i=0; i<nvar; i++) {
      fprintf(src,"_jac_[%d] = ",n*nvar+i);
      node = cptree(dffeqns[n]);
      for (j=nimvar; j>0; j--)
	node = pasteim(node,j-1);
      node = cltree(deriv(node,varnms[i]));
      for (j=0; j<nimvar; j++) node = cutim(node,j);
      for (j=0; j<nimvar; j++) node = cutim(node,j);
      if (jacob) {
	prtree(node,src); 
	fp(";");
      }
      else 
	fp("0.;");
      node = rmtree(node);
    }
    fprintf(src,"if (!forward_opt)\n  for (_i_=0; _i_<%d; _i_++)\n",nvar*nvar); 
    fp("    _jac_[_i_] = - _jac_[_i_];"); 
  }
  fp("}\n");
  return 1;
}

int wrmain() {
  fp("int main(int argc, char *argv[])\n{");  
  if (jacob) fp("  jacob_opt = 1;");
  else fp("  jacob_opt = 0;");
  fp("  Initialize(argc, argv);");
  fp("  GrindMainLoop(1);");
  fp("  return 0;");
  fp("}\n");
  return 1;
}
