#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <signal.h>
#include <setjmp.h>
#include <string.h>
#include <unistd.h>

/* #if USE_GET_LINE */
#include "../gifts/getline.h"
/* #endif */

#include "grind.h"

extern FILE *cinput;
extern char uline[];
extern char cline[];
extern char *cbegin;
extern char *cend;
extern int margc;
extern char *margv[];
extern char *modelname;

int readLineInteractive() {
  static boolean first = true;
  int c = '\0', i;
  char *new_line;
  char prompt[] = "GRIND> ";

  if(first) {
    first = false;
    gl_tab_completion(0);
    if (LINE_LENGTH >= GL_BUF_SIZE) {
      printf("incompatible buffer sizes; change readLine()\n");
      exit(1);
    }
  }

  i = 0;
  new_line = getline2(prompt);
  if (GL_OK == gl_get_result()) {
    while ( (c=new_line[i]) != '\n' && c != '\0')
      uline[i++] = c;
    if (i)
      gl_histadd(new_line);
  }
  uline[i] = '\0';
  if (GL_EOF == gl_get_result())
    return EOF;
  else
    return i;  
}

int readLine(boolean interactive) {
  int c = '\0', i, lim;

  if (isatty(0) && isatty(1) && interactive && USE_GET_LINE)
    return readLineInteractive();

  i = 0;
  lim = LINE_LENGTH;
  while (--lim > 0 && (c = getc(cinput)) != EOF && c != '\n') {
    uline[i++] = c;
  }
  if (!lim)
    gr_warning("input exceeds buffer limit (%d) in readLine()\n", LINE_LENGTH);
  uline[i] = '\0';
  if (EOF == c)
    return EOF;
  else
    return i;
}

int ParseArgs() {
  /*  extract the next command from uline, parse it into subsequent args */
  /*  and store these in cline; each ending with a '\0'  */
  /*  and the next argv[i] pointing to its beginning.  */
  /*  Note that spaces, comma's and equal-signs are ignored and skipped */

  int c;
  boolean writing = false;
  /*  at this point 'cend' points to the index of 'uline'  */
  /*  where the next command starts */
  char *pline = cend,*pcline=cline;

  /*  reset margc, set cbegin to the beginning of the new command */
  margc = 0;
  cbegin = cend;
  /*  copy uline to cline until end of this command */
  while ((c = *pline++) != '\0' && c != ';') {
    /*  check for assignment */
    if (c == '=') {
      if (writing) {
	/*  former comand is completely read; end it */
	*pcline++ = '\0';
      }
      /*  '=' becomes next argv */
      margv[margc++] = pcline;
      *pcline++ = c;
      /*  end this argv and stop writing */
      *pcline++ = '\0';
      writing = false;
    }
    /*  check for delimiters of commands */
    else if (c == ' ' || c == '\t' || c == ',') {
      if (writing) {
	/*  here a command is completely read */
        *pcline++ = '\0';
	writing = false;
      }
    }
    else {
      if (writing)
	*pcline++ = c;
      else {
	/*  start a new arg */
	writing = true;
	margv[margc++] = pcline;
	*pcline++ = c;
      }
    }
  }
  *pcline++ = '\0';
  /*  set pointer to end of this command in 'uline' */
  cend=pline;
  /* } */
  /*  return whether we read up to the end of the line (ie, when c == '\0') */
  return c;
}

int GetCommand(boolean interactive) {
  char c = '\0';
  static int first = 1;
  int ret;
  static FILE *logfile;
  static char *logfilename = NULL;

  /*  sigsetjmp(return_to_top_level, 1); */

  if (logfile_opt) {
    if (first) {
      first = false;
      logfilename = ccalloc(strlen(modelname) + 5, 
			    "'logfilename' in GetCommend()");
      strcpy(logfilename, modelname);
      strcat(logfilename, ".log");
      if ((logfile = fopen(logfilename,"w")) == NULL)
	gr_warning("unable to open logfile %s\n", logfilename);
    }
    else if ((logfile = fopen(logfilename,"a")) == NULL)
      gr_warning("unable to open logfile %s\n", logfilename);
  }

  /* get a new line from user, store in 'uline' */
  if ((ret = readLine(interactive)) == 0) {
    if (logfile_opt)
      fclose(logfile);
    /*  if nothing was read */
    return 0;
  }
  if (logfile_opt && logfile) {
    fprintf(logfile, "%s\n", uline);
    fclose(logfile);
  }

  /*  set some pointers to the beginning of the just read input */
  cbegin = cend = uline;
  /*  extract commands from uline, and execute them, one command at a time */
  do {
    /*  extract next command from 'uline' */
    c = ParseArgs();
    /*  if there are commands, execute them */
    if (margc)
      Exec(margc,margv);
  }
  while (c != '\0');
  if (EOF == ret)
    return ret;
  return 1;
}
